import os
import re
from mysqlx import db
from jinja2 import Template
from mysqlx.model import SqlModel
try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET

_REGEX = '{%|{{|}}|%}'
_SQL_CONTAINER = dict()


def init_db(user, password, database, host='127.0.0.1', port=3306, use_unicode=True, mapper_path='mapper', **kw):
    _load_sql(mapper_path)
    db.init_db(user, password, database, host, port, use_unicode, **kw)


def insert(table, **kw):
    return db.insert(table, **kw)


def execute(sql_id, *args):
    sql = get_sql(sql_id)
    return db.execute(sql, *args)


def batch_execute(sql_id, args: list):
    sql = get_sql(sql_id)
    return db.batch_execute(sql, args)


def get(sql_id, *args):
    sql = get_sql(sql_id)
    return db.get(sql, *args)


def select_one(sql_id, *args):
    sql = get_sql(sql_id)
    return db.select_one(sql, *args)


def select(sql_id, *args):
    sql = get_sql(sql_id)
    return db.select(sql, *args)


def named_execute(sql_id, **kwargs):
    sql = get_sql(sql_id, **kwargs)
    return db.named_execute(sql, **kwargs)


def named_get(sql_id, **kwargs):
    sql = get_sql(sql_id, **kwargs)
    return db.named_get(sql, **kwargs)


def named_select_one(sql_id, **kwargs):
    sql = get_sql(sql_id, **kwargs)
    return db.named_select_one(sql, **kwargs)


def named_select(sql_id, **kwargs):
    sql = get_sql(sql_id, **kwargs)
    return db.named_select(sql, **kwargs)


def get_connection():
    return db.get_connection()


def _get_path(path):
    if path.startswith("../"):
        rpath = ''.join(re.findall("../", path))
        os.chdir(rpath)
        path = path[len(rpath):]
    elif path.startswith("./"):
        path = path[2:]
    return os.path.join(os.getcwd(), path)


def _load_sql(path):
    if not os.path.isabs(path):
        path = _get_path(path)

    for f in os.listdir(path):
        file = os.path.join(path, f)
        if os.path.isfile(file) and f.endswith(".xml"):
            _read_mapper(file)
        elif os.path.isdir(file):
            _load_sql(file)


def _read_mapper(file):
    global _SQL_CONTAINER
    tree = ET.parse(file)
    root = tree.getroot()
    namespace = root.attrib.get('namespace', '')
    for child in root:
        sql_id = namespace + "." + child.attrib.get('id')
        include = child.attrib.get('include')
        sql = child.text
        if re.search(_REGEX, sql) or include:
            _SQL_CONTAINER[sql_id] = SqlModel(sql=Template(sql), dynamic=True, include=include)
        else:
            _SQL_CONTAINER[sql_id] = SqlModel(sql=sql)


def get_sql(sql_id, **kwargs):
    sql_model = _get_sql_model(sql_id)
    include = sql_model.include
    if include:
        include_sql_id = sql_id[:sql_id.index(".")+1] + include
        kwargs[include] = get_sql(include_sql_id, **kwargs)
    return sql_model.sql.render(**kwargs) if sql_model.dynamic else sql_model.sql


def _get_sql_model(sql_id):
    global _SQL_CONTAINER
    return _SQL_CONTAINER[sql_id]




import logging
import threading


class DBCtx(threading.local):
    """
    Thread local object that holds connection info.
    """

    def __init__(self, connect):
        self.connect = connect
        self.connection = None
        self.transactions = 0

    def is_init(self):
        return self.connection is not None

    def init(self):
        self.transactions = 0
        self.connection = self.connect()
        logging.debug('use connection <%s>...' % hex(id(self.connection)))

    def release(self):
        if self.connection:
            self.connection.close()
            logging.debug('release connection <%s>...' % hex(id(self.connection)))
            self.connection = None

    def cursor(self):
        """
        Return cursor
        """
        return self.connection.cursor()

    def statement(self, sql):
        """
        Return statement
        """
        return self.connection.statement(sql)


class ConnectionCtx(object):
    """
    ConnectionCtx object that can open and close connection context. _ConnectionCtx object can be nested and only the most
    outer connection has effect.
    with connection():
        pass
        with connection():
            pass
    """

    def __init__(self, db_ctx):
        self.db_ctx = db_ctx

    def __enter__(self):
        self.should_cleanup = False
        if not self.db_ctx.is_init():
            self.db_ctx.init()
            self.should_cleanup = True
        return self

    def __exit__(self, exctype, excvalue, traceback):
        if self.should_cleanup:
            self.db_ctx.release()


class TransactionCtx(object):
    """
    TransactionCtx object that can handle transactions.
    with TransactionCtx():
        pass
    """

    def __init__(self, db_ctx):
        self.db_ctx = db_ctx

    def __enter__(self):
        self.should_close_conn = False
        if not self.db_ctx.is_init():
            # needs open a connection first:
            self.db_ctx.init()
            self.should_close_conn = True
        self.db_ctx.transactions = self.db_ctx.transactions + 1
        logging.info('begin transaction...' if self.db_ctx.transactions == 1 else 'join current transaction...')
        return self

    def __exit__(self, exctype, excvalue, traceback):
        self.db_ctx.transactions -= 1
        try:
            if self.db_ctx.transactions == 0:
                if exctype is None:
                    self.commit()
                else:
                    self.rollback()
        finally:
            if self.should_close_conn:
                self.db_ctx.release()

    def commit(self):
        logging.info('commit transaction...')
        try:
            self.db_ctx.connection.commit()
            logging.info('commit ok.')
        except Exception:
            logging.warning('commit failed. try rollback...')
            self.db_ctx.connection.rollback()
            logging.warning('rollback ok.')
            raise

    def rollback(self):
        logging.warning('rollback transaction...')
        self.db_ctx.connection.rollback()
        logging.info('rollback ok.')


class DBError(Exception):
    pass


class MultiColumnsError(DBError):
    pass


class Dict(dict):
    """
    Simple dict but support access as x.y style.
    >>> d1 = Dict()
    >>> d1['x'] = 100
    >>> d1.x
    100
    >>> d1.y = 200
    >>> d1['y']
    200
    >>> d2 = Dict(a=1, b=2, c='3')
    >>> d2.c
    '3'
    >>> d2['empty']
    Traceback (most recent call last):
        ...
    KeyError: 'empty'
    >>> d2.empty
    Traceback (most recent call last):
        ...
    AttributeError: 'Dict' object has no attribute 'empty'
    >>> d3 = Dict(('a', 'b', 'c'), (1, 2, 3))
    >>> d3.a
    1
    >>> d3.b
    2
    >>> d3.c
    3
    """

    def __init__(self, names=(), values=(), **kw):
        super(Dict, self).__init__(**kw)
        for k, v in zip(names, values):
            self[k] = v

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(r"'Dict' object has no attribute '%s'" % key)

    def __setattr__(self, key, value):
        self[key] = value


class SqlModel:

    def __init__(self, sql, dynamic=False, include=None):
        self.sql = sql
        self.dynamic = dynamic
        self.include = include




import re, time, functools, logging
from mysqlx.model import DBCtx, ConnectionCtx, Dict, MultiColumnsError, TransactionCtx

# thread-local context:
_DB_CTX = None
_REGEX = r':[\w|\d]*'


def init_db(user, password, database, host='127.0.0.1', port=3306, use_unicode=True, **kwargs):
    from mysql.connector import connect
    global _DB_CTX
    if ('pool_size' in kwargs or 'pool_reset_session' in kwargs) and 'pool_name' not in kwargs:
        kwargs['pool_name'] = database

    kwargs['user'] = user
    kwargs['password'] = password
    kwargs['database'] = database
    kwargs['host'] = host
    kwargs['port'] = port
    kwargs['use_unicode'] = use_unicode
    _DB_CTX = DBCtx(lambda: connect(**kwargs))
    logging.info('Init db engine <%s> ok.' % hex(id(_DB_CTX)))


def connection():
    """
    Return _ConnectionCtx object that can be used by 'with' statement:
    with connection():
        pass
    """
    global _DB_CTX
    return ConnectionCtx(_DB_CTX)


def with_connection(func):
    """
    Decorator for reuse connection.
    @with_connection
    def foo(*args, **kw):
        f1()
        f2()
        f3()
    """

    global _DB_CTX

    @functools.wraps(func)
    def _wrapper(*args, **kw):
        with ConnectionCtx(_DB_CTX):
            return func(*args, **kw)

    return _wrapper


def transaction():
    """
    Create a transaction object so can use with statement:
    with transaction():
        pass
    >>> def update_profile(id, name, rollback):
    ...     u = dict(id=id, name=name, email='%s@test.org' % name, passwd=name, last_modified=time.time())
    ...     insert('user', **u)
    ...     r = update('update user set passwd=? where id=?', name.upper(), id)
    ...     if rollback:
    ...         raise StandardError('will cause rollback...')
    >>> with transaction():
    ...     update_profile(900301, 'Python', False)
    >>> select_one('select * from user where id=?', 900301).name
    u'Python'
    >>> with transaction():
    ...     update_profile(900302, 'Ruby', True)
    Traceback (most recent call last):
      ...
    StandardError: will cause rollback...
    >>> select('select * from user where id=?', 900302)
    []
    """
    global _DB_CTX
    return TransactionCtx(_DB_CTX)


def with_transaction(func):
    """
    A decorator that makes function around transaction.
    >>> @with_transaction
    ... def update_profile(id, name, rollback):
    ...     u = dict(id=id, name=name, email='%s@test.org' % name, passwd=name, last_modified=time.time())
    ...     insert('user', **u)
    ...     r = update('update user set passwd=? where id=?', name.upper(), id)
    ...     if rollback:
    ...         raise StandardError('will cause rollback...')
    >>> update_profile(8080, 'Julia', False)
    >>> select_one('select * from user where id=?', 8080).passwd
    u'JULIA'
    >>> update_profile(9090, 'Robert', True)
    Traceback (most recent call last):
      ...
    StandardError: will cause rollback...
    >>> select('select * from user where id=?', 9090)
    []
    """
    global _DB_CTX

    @functools.wraps(func)
    def _wrapper(*args, **kw):
        _start = time.time()
        with TransactionCtx(_DB_CTX):
            return func(*args, **kw)
        _profiling(_start)

    return _wrapper


def _select(sql, first, *args):
    """execute select SQL and return unique result or list results."""
    global _DB_CTX
    cursor = None
    logging.info('SQL: %s, ARGS: %s' % (sql, args))
    sql = sql.replace('?', '%s')
    try:
        cursor = _DB_CTX.connection.cursor()
        cursor.execute(sql, args)
        if cursor.description:
            names = [x[0] for x in cursor.description]
        if first:
            values = cursor.fetchone()
            if not values:
                return None
            return Dict(names, values)
        return [Dict(names, x) for x in cursor.fetchall()]
    finally:
        if cursor:
            cursor.close()


@with_connection
def get(sql, *args):
    """
    Execute select SQL and expected one int and only one int result.
    MultiColumnsError: Expect only one column.
    """
    global _DB_CTX
    cursor = None
    logging.info('SQL: %s, ARGS: %s' % (sql, args))
    sql = sql.replace('?', '%s')
    try:
        cursor = _DB_CTX.connection.cursor()
        cursor.execute(sql, args)
        values = cursor.fetchone()
        if len(values) == 1:
            return values[0]
        else:
            raise MultiColumnsError('Expect only one column.')
    finally:
        if cursor:
            cursor.close()


@with_connection
def select_one(sql, *args):
    """
    Execute select SQL and expected one result.
    If no result found, return None.
    If multiple results found, the first one returned.
    >>> u1 = dict(id=100, name='Alice', email='alice@test.org', passwd='ABC-12345', last_modified=time.time())
    >>> u2 = dict(id=101, name='Sarah', email='sarah@test.org', passwd='ABC-12345', last_modified=time.time())
    >>> insert('user', **u1)
    1
    >>> insert('user', **u2)
    1
    >>> u = select_one('select * from user where id=?', 100)
    >>> u.name
    u'Alice'
    >>> select_one('select * from user where email=?', 'abc@email.com')
    >>> u2 = select_one('select * from user where passwd=? order by email', 'ABC-12345')
    >>> u2.name
    u'Alice'
    """
    return _select(sql, True, *args)


@with_connection
def select(sql, *args):
    """
    Execute select SQL and return list or empty list if no result.
    >>> u1 = dict(id=200, name='Wall.E', email='wall.e@test.org', passwd='back-to-earth', last_modified=time.time())
    >>> u2 = dict(id=201, name='Eva', email='eva@test.org', passwd='back-to-earth', last_modified=time.time())
    >>> insert('user', **u1)
    1
    >>> insert('user', **u2)
    1
    >>> L = select('select * from user where id=?', 900900900)
    >>> L
    []
    >>> L = select('select * from user where id=?', 200)
    >>> L[0].email
    u'wall.e@test.org'
    >>> L = select('select * from user where passwd=? order by id desc', 'back-to-earth')
    >>> L[0].name
    u'Eva'
    >>> L[1].name
    u'Wall.E'
    """
    return _select(sql, False, *args)


@with_connection
def _execute(sql, *args):
    global _DB_CTX
    cursor = None
    logging.info('SQL: %s, ARGS: %s' % (sql, args))
    sql = sql.replace('?', '%s')
    try:
        cursor = _DB_CTX.connection.cursor()
        cursor.execute(sql, args)
        r = cursor.rowcount
        if _DB_CTX.transactions == 0:
            _DB_CTX.connection.commit()
        return r
    finally:
        if cursor:
            cursor.close()


def insert(table, **kw):
    """
    Execute insert SQL.
    IntegrityError: 1062 (23000): Duplicate entry '2000' for key 'PRIMARY'
    """
    cols, args = zip(*kw.items())
    sql = 'insert into `%s` (%s) values (%s)' % (table, ','.join(['`%s`' % col for col in cols]), ','.join(['?' for i in range(len(cols))]))
    return _execute(sql, *args)


def execute(sql, *args):
    """
    Execute SQL.
    """
    return _execute(sql, *args)


@with_connection
def batch_execute(sql, args: list):
    global _DB_CTX
    cursor = None
    logging.info('SQL: %s, ARGS: %s' % (sql, args))
    sql = sql.replace('?', '%s')
    try:
        cursor = _DB_CTX.connection.cursor()
        cursor.executemany(sql, args)
        r = cursor.rowcount
        if _DB_CTX.transactions == 0:
            _DB_CTX.connection.commit()
        return r
    finally:
        if cursor:
            cursor.close()


def named_execute(sql, **kwargs):
    sql, args = _get_named_sql_args(sql, **kwargs)
    return execute(sql, *args)


def named_get(sql, **kwargs):
    sql, args = _get_named_sql_args(sql, **kwargs)
    return get(sql, *args)


def named_select_one(sql, **kwargs):
    sql, args = _get_named_sql_args(sql, **kwargs)
    return select_one(sql, *args)


def named_select(sql, **kwargs):
    sql, args = _get_named_sql_args(sql, **kwargs)
    return select(sql, *args)


def get_connection():
    global _DB_CTX
    if not _DB_CTX.is_init():
        _DB_CTX.init()
    return _DB_CTX.connection


def _get_named_sql_args(sql, **kwargs):
    args = [kwargs[r[1:]] for r in re.findall(_REGEX, sql)]
    return re.sub(_REGEX, '?', sql), args



from setuptools import setup, find_packages

# INSTALL_PACKAGES = open(path.join(DIR, 'requirements.txt')).read().splitlines()

setup(
    name='mysqlx',
    packages=find_packages(),
    description="mysqlx is a simple python sql executor for MySQL like iBatis.",
    long_description_content_type='text/markdown',
    install_requires=[
        'Jinja2>=3.0.3',
        'mysql-connector-python>=8.0.27',
    ],
    version='1.1.1',
    url='https://gitee.com/summry/mysqlx/mysqlx',
    author='summry',
    author_email='xiazhongbiao@126.com',
    keywords=['sql', 'MySQL', 'iBatis', 'MyBatis', 'python'],
    package_data={
        # include json and txt files
        '': ['*.rst', '*.txt'],
    },
    include_package_data=True,
    python_requires='>=3.6.0',
    zip_safe=False
)

